Index: dodsdataset2.cpp
===================================================================
--- dodsdataset2.cpp	(revision 24933)
+++ dodsdataset2.cpp	(working copy)
@@ -74,6 +74,7 @@
 
 CPL_C_START
 void GDALRegister_DODS(void);
+char **Direct_Product(int, char **);
 CPL_C_END
 
 /** Attribute names used to encode geo-referencing information. Note that
@@ -231,8 +232,8 @@
     char            **CollectBandsFromDDSVar( string, char ** );
     char            **ParseBandsFromURL( string );
 
-    void            HarvestDAS();
-    static void     HarvestMetadata( GDALMajorObject *, AttrTable * );
+    void            HarvestDAS( string, string , string );
+    static void     HarvestMetadata( GDALMajorObject *, AttrTable *, string, string );
     void            HarvestMaps( string oVarName, string oCE );
     
     friend class DODSRasterBand;
@@ -284,7 +285,7 @@
     int    bNoDataSet;
     double dfNoDataValue;
 
-    void            HarvestDAS();
+    void            HarvestDAS(string);
 
 public:
     DODSRasterBand( DODSDataset *poDS, string oVarName, string oCE, 
@@ -439,7 +440,6 @@
             + raw_constraint.substr( y_off + y_len, x_off - y_off - y_len )
             + x_constraint
             + raw_constraint.substr( x_off + x_len );
-
     return final_constraint;
 }
 
@@ -468,12 +468,11 @@
 
     for( v_i = poDDS->var_begin(); v_i != poDDS->var_end(); v_i++ )
     {
-        BaseType *poVar = *v_i;
-        papszResultList = CollectBandsFromDDSVar( poVar->name(), 
+		BaseType *poVar = *v_i;
+		papszResultList = CollectBandsFromDDSVar( poVar->name(), 
                                                   papszResultList );
     }
-
-    return papszResultList;
+	return papszResultList;
 }
 
 /************************************************************************/
@@ -488,12 +487,27 @@
 
 char **DODSDataset::CollectBandsFromDDSVar( string oVarName, 
                                             char **papszResultList )
-
 {
     Array *poArray;
     Grid *poGrid = NULL;
-    
+	string::size_type index;
+	string oConstraint;
+
 /* -------------------------------------------------------------------- */
+/*	 if oVarName is given as "v_goce[1][y][x]",							*/			
+/*	  it is devided into oConstraint and oVarName which represent		*/	
+/*	  the dimensions and variableName.									*/		
+/*	   e.g.) oConstraint = "[1][y][x]" and oVarName = "v_goce"			*/
+/* -------------------------------------------------------------------- */
+
+	index = oVarName.find("[");
+	if( index != string::npos )
+	{
+		oConstraint = oVarName.substr(index);
+        oVarName = oVarName.substr(0,index);
+	}
+
+/* -------------------------------------------------------------------- */
 /*      Is this a grid or array?                                        */
 /* -------------------------------------------------------------------- */
     BaseType *poVar = get_variable( GetDDS(), oVarName );
@@ -512,92 +526,225 @@
         return papszResultList;
 
 /* -------------------------------------------------------------------- */
-/*      Eventually we will want to support arrays with more than two    */
-/*      dimensions ... but not quite yet.                               */
+/*      Support arrays with two and more dimensions                     */
 /* -------------------------------------------------------------------- */
-    if( poArray->dimensions() != 2 )
-        return papszResultList;
+	if( poArray->dimensions() < 2)    	
+		return papszResultList;
 
 /* -------------------------------------------------------------------- */
-/*      Get the dimension information for this variable.                */
+/*      Get the dimension information for all variable.                 */
 /* -------------------------------------------------------------------- */
-    Array::Dim_iter dim1 = poArray->dim_begin() + 0;
-    Array::Dim_iter dim2 = poArray->dim_begin() + 1;
-    
-    int nDim1Size = poArray->dimension_size( dim1 );
-    int nDim2Size = poArray->dimension_size( dim2 );
-    
-    if( nDim1Size == 1 || nDim2Size == 1 )
-        return papszResultList;
+	Array::Dim_iter *poDim;
+	int *poDimSize;
+	poDim = (Array::Dim_iter *)CPLMalloc(
+					poArray->dimensions() * sizeof( Array::Dim_iter ));
+	poDimSize = (int *)CPLMalloc( poArray->dimensions() * sizeof(int) );
 
+	// set pointers and sizes of each dimension
+	for(int i = 0; i < (signed int) poArray->dimensions(); i++ )
+	{
+		poDim[i] = poArray->dim_begin() + i;
+		poDimSize[i] = poArray->dimension_size(poDim[i]);
+	}
+
+	if(poDimSize[poArray->dimensions()-1] == 1 || 
+			poDimSize[poArray->dimensions()-2] == 1) 
+		return papszResultList;
+	
+	CPLFree( poDimSize );
+
 /* -------------------------------------------------------------------- */
 /*      Try to guess which is x and y.                                  */
+/*			NB : if "x" and "y" are included in dim_name,               */
+/*               x and y axses are guessed.								*/
+/*               But it is not safe because of one letter.				*/
+/*               Should be fixed.										*/
 /* -------------------------------------------------------------------- */
-    string dim1_name = poArray->dimension_name( dim1 );
-    string dim2_name = poArray->dimension_name( dim2 );
-    int iXDim=-1, iYDim=-1;
+	int *poAxesOrder;
+	string dim_1_name;
+    string dim_2_name;
+	string 	tmpSubString;
+	
+	char dim_name[][3][20] = {{"east", "lon", "x"},
+							  {"north", "lat", "y"}};
 
-    if( dim1_name == "easting" && dim2_name == "northing" )
-    {
-        iXDim = 0;
-        iYDim = 1;
-    }
-    else if( dim1_name == "easting" && dim2_name == "northing" )
-    {
-        iXDim = 1;
-        iYDim = 0;
-    }
-    else if( EQUALN(dim1_name.c_str(),"lat",3) 
-             && EQUALN(dim2_name.c_str(),"lon",3) )
-    {
-        iXDim = 0;
-        iYDim = 1;
-    }
-    else if( EQUALN(dim1_name.c_str(),"lon",3) 
-             && EQUALN(dim2_name.c_str(),"lat",3) )
-    {
-        iXDim = 1;
-        iYDim = 0;
-    }
-    else 
-    {
-        iYDim = 0;
-        iXDim = 1;
-    }
+	poAxesOrder = (int *)CPLMalloc( poArray->dimensions() * sizeof(int) );
 
 /* -------------------------------------------------------------------- */
+/*	 fetch coordinate names of the last and the second last coordinates.*/
+/* 	  if x and y axes are given by oConstraint (in case of "else"),     */
+/*	  set the variable names to dim_1_name and dim_2_name.              */
+/*	   e.g.) if oConstraint=[y][x]                                      */
+/*					then dim_1_name = "x" and dim_2_name ="y"           */
+/*	    	 if oConstraint=[xc][yc]                                    */
+/*					then dim_1_name = "yc" and dim_2_name ="xc"         */
+/* -------------------------------------------------------------------- */
+	if (oConstraint == "")
+	{
+	    dim_2_name = poArray->dimension_name(poDim[poArray->dimensions()-2]);
+		dim_1_name = poArray->dimension_name(poDim[poArray->dimensions()-1]);
+	}
+	else
+	{
+		index = oConstraint.find("[");
+		tmpSubString = oConstraint.substr(index);
+
+		index = tmpSubString.find_last_of("]");
+		tmpSubString = tmpSubString.substr(0, index);
+		index = tmpSubString.find_last_of("[");
+		dim_1_name = tmpSubString.substr(index+1);
+		
+		index = tmpSubString.find_last_of("]");
+		tmpSubString = tmpSubString.substr(0, index);
+		index = tmpSubString.find_last_of("[");
+		dim_2_name = tmpSubString.substr(index+1);
+	}
+/* -------------------------------------------------------------------- */
+/*	 the dimension is regarded as "[0][1][2][3]..."                     */
+/* -------------------------------------------------------------------- */
+	for (int i = 0; i < (signed int)poArray->dimensions(); i++)
+		poAxesOrder[i] = i;
+
+/* -------------------------------------------------------------------- */
+/*	if [east..][north..], [lon..][lat..] or [x..][y..],                 */
+/*	  then switch the last two variables of poAxesOrder                 */
+/*	   e.g. [0][1][2][3] is changed as [0][1][3][2]                     */
+/* -------------------------------------------------------------------- */
+	for (int i = 0; i < 3; i++)
+	{
+		if ( EQUALN( dim_2_name.c_str(), dim_name[0][i], 
+						strlen( dim_name[0][i] )) 
+             && EQUALN( dim_1_name.c_str(), dim_name[1][i], 
+						strlen( dim_name[1][i] )))
+		{
+			int tmp = poAxesOrder[poArray->dimensions()-2];
+			poAxesOrder[poArray->dimensions()-2] = poAxesOrder[poArray->dimensions()-1];
+			poAxesOrder[poArray->dimensions()-1] = tmp;
+			break;
+		}
+	}
+	CPLFree( poDim );
+
+/* -------------------------------------------------------------------- */
 /*      Does this match the established dimension?                      */
 /* -------------------------------------------------------------------- */
-    Array::Dim_iter dimx = poArray->dim_begin() + iXDim;
-    Array::Dim_iter dimy = poArray->dim_begin() + iYDim;
+	int *poRasterSize;
+	Array::Dim_iter *poDimAxis; 
 
+	poDimAxis = ( Array::Dim_iter * )CPLMalloc( 
+					poArray->dimensions() * sizeof(Array::Dim_iter) );
+	poRasterSize = (int *)CPLMalloc(poArray->dimensions() * sizeof(int) );
+
+/* -------------------------------------------------------------------- */
+/*		get size of each dimension.										*/		
+/*		e.g.) if [time][y][x] = [5][500][800]							*/
+/*				then poRasterSize[0]=5, poRasterSize[1]=500,			*/
+/*										poRasterSize[2]=800				*/
+/*			  if [time][x][y] = [5][500][800]							*/
+/*				then poRasterSize[0]=5, poRasterSize[1]=800,			*/
+/*										poRasterSize[2]=500				*/	
+/* -------------------------------------------------------------------- */
+	for(int i = 0; i < (signed int) poArray->dimensions(); i++)
+	{
+		poDimAxis[i] = poArray->dim_begin() + poAxesOrder[i];
+		poRasterSize[i] = poArray->dimension_size(poDimAxis[i]);
+	}
+
+	// if it is the first band, set the rastersize
     if( nRasterXSize == 0 && nRasterYSize == 0 )
-    {
-        nRasterXSize = poArray->dimension_size( dimx );
-        nRasterYSize = poArray->dimension_size( dimy );
-    }
+	{
+		nRasterXSize = poRasterSize[poArray->dimensions()-1];
+		nRasterYSize = poRasterSize[poArray->dimensions()-2];
+	}
 
-    if( nRasterXSize != poArray->dimension_size( dimx )
-        || nRasterYSize != poArray->dimension_size( dimy ) )
+    if( nRasterXSize != poRasterSize[poArray->dimensions()-1]
+        || nRasterYSize != poRasterSize[poArray->dimensions()-2] )
         return papszResultList;
 
+	CPLFree( poDimAxis );
+
 /* -------------------------------------------------------------------- */
 /*      OK, we have an acceptable candidate!                            */
 /* -------------------------------------------------------------------- */
-    string oConstraint;
+	char **papszOConstraintList = NULL;
+	int combNum = 1;
+/* -------------------------------------------------------------------- */
+/*	 variable for setting "dimension of variable" and "variableName".   */
+/*	 It is used in DODSDataset::HarvestMetadata()                       */
+/*	 to add band information in metadata                                */
+/* -------------------------------------------------------------------- */
+	string bandCoordinate;
+	if (oConstraint =="")
+	{
+		if (poAxesOrder[poArray->dimensions()-1] == (signed int)poArray->dimensions()-1 
+			&& poAxesOrder[poArray->dimensions()-2] == (signed int)poArray->dimensions()-2)
+			oConstraint = "[y][x]";
+		else if (poAxesOrder[poArray->dimensions()-1] == (signed int)poArray->dimensions()-2 
+			&& poAxesOrder[poArray->dimensions()-2] == (signed int)poArray->dimensions()-1)
+			oConstraint = "[x][y]";
+		else
+			return papszResultList;
+/* -------------------------------------------------------------------- */
+/*		 Create papszOConstraintList which has "[y][x]" or "[x][y]"     */
+/* -------------------------------------------------------------------- */
+		papszOConstraintList = CSLAddString( papszOConstraintList, 
+											 oConstraint.c_str() );
+/* -------------------------------------------------------------------- */
+/*		 If the dimension is three and more,                            */
+/*		 create papszOConstraintList which has all combinations.        */
+/*		 e.g. if [time][depth][y][x], time = 1,2,3 and depth = 1,2      */
+/*		      then [1][1][y][x], [1][2][y][x], [2][1][y][x],            */
+/*                                        ... ,[3][2][y][x]             */
+/* -------------------------------------------------------------------- */
+		for (int i = 0; i < (signed int)poArray->dimensions()-2; i++)
+		{
+			papszOConstraintList = 
+				Direct_Product( poRasterSize[poArray->dimensions()-3-i], 
+								papszOConstraintList);
+			combNum *= poRasterSize[poArray->dimensions()-3-i];
+		}
+	}
+	else
+	{
+/* -------------------------------------------------------------------- */
+/*		if a band is specified by users,                                */
+/*        set it into papszOConstraintList.                             */
+/*		 e.g. [3][y][x]                                                 */
+/* -------------------------------------------------------------------- */
+		papszOConstraintList = CSLAddString( papszOConstraintList, 
+													oConstraint.c_str());
+	}
+/* -------------------------------------------------------------------- */
+/*	Create a string which has "?" + "[ variable name]".                 */
+/*	 e.g. "?[latitude][longitude]", "?[time][lat][lon]", ...            */
+/* -------------------------------------------------------------------- */
+	string coordName = "?";
+	char chr[20]; 	
+	for (int i = 0; i < (signed int)poArray->dimensions(); i++)
+	{
+		sprintf(chr,"%s", 
+			poArray->dimension_name(poArray->dim_begin() + poAxesOrder[i]).c_str());
+		coordName += "[" ;
+		coordName += string(chr);
+		coordName += "]";
+	}
 
-    if( iXDim == 0 && iYDim == 1 )
-        oConstraint = "[x][y]";
-    else if( iXDim == 1 && iYDim == 0 )
-        oConstraint = "[y][x]";
-    else
-        return papszResultList;
-
-    papszResultList = CSLAddString( papszResultList, 
-                                    poVar->name().c_str() );
-    papszResultList = CSLAddString( papszResultList, 
-                                    oConstraint.c_str() );
-
+	for (int i=0; i < combNum ; i++)
+	{
+		papszResultList = CSLAddString( papszResultList, 
+									poVar->name().c_str());
+/* -------------------------------------------------------------------- */
+/*		make a string as ' dimension of variable + "?" + variableName'  */
+/*		e.g. "[y][x]?[lat][lon]",                                       */
+/*		     "[1][yc][xc]?[time][latitude][longitude]", ...             */
+/* -------------------------------------------------------------------- */
+		bandCoordinate = papszOConstraintList[i] + coordName;
+		papszResultList = CSLAddString( papszResultList, 
+									bandCoordinate.c_str());
+	}
+	
+	CPLFree( poRasterSize );
+	CPLFree( poAxesOrder );
     return papszResultList;
 }
 
@@ -609,7 +756,7 @@
 
 {
     char **papszResultList = NULL;
-    char **papszVars = CSLTokenizeString2( oVarList.c_str(), ",", 0 );
+	char **papszVars = CSLTokenizeString2( oVarList.c_str(), ",", 0 );
     int i;
 
     for( i = 0; papszVars != NULL && papszVars[i] != NULL; i++ )
@@ -620,33 +767,9 @@
 /* -------------------------------------------------------------------- */
 /*      Split into a varname and constraint equation.                   */
 /* -------------------------------------------------------------------- */
-        char *pszCEStart = strstr(papszVars[i],"[");
-
-        // If we have no constraints we will have to try to guess
-        // reasonable values from the DDS.  In fact, we might end up
-        // deriving multiple bands from one variable in this case.
-        if( pszCEStart == NULL )
-        {
-            oVarName = papszVars[i];
-
-            papszResultList = 
-                CollectBandsFromDDSVar( oVarName, papszResultList );
-        }
-        else
-        {
-            oCE = pszCEStart;
-            *pszCEStart = '\0';
-            oVarName = papszVars[i];
-
-            // Eventually we should consider supporting a [band] keyword
-            // to select a constraint variable that should be used to
-            // identify a band dimension ... but not for now. 
-
-            papszResultList = CSLAddString( papszResultList,
-                                            oVarName.c_str() );
-            papszResultList = CSLAddString( papszResultList, 
-                                            oCE.c_str() );
-        }
+		oVarName = papszVars[i];
+        papszResultList = 
+            CollectBandsFromDDSVar( oVarName, papszResultList );
     }
 
     CSLDestroy(papszVars);
@@ -662,56 +785,113 @@
 /************************************************************************/
 
 void DODSDataset::HarvestMetadata( GDALMajorObject *poTarget, 
-                                   AttrTable *poSrcTable )
-
+                                   AttrTable *poSrcTable, 
+								   string containerName, string coordName)
 {
     if( poSrcTable == NULL )
         return;
 
 /* -------------------------------------------------------------------- */
-/*      Find Metadata container.                                        */
+/*      Find Metadata container. if not, take the argunent poSrcTable   */
 /* -------------------------------------------------------------------- */
     AttrTable *poMDTable = poSrcTable->find_container("Metadata");
-    
-    if( poMDTable == NULL )
-        return;
+	if( poMDTable == NULL )
+		poMDTable = poSrcTable;
 
 /* -------------------------------------------------------------------- */
 /*      Collect each data item from it.                                 */
 /* -------------------------------------------------------------------- */
     AttrTable::Attr_iter dv_i;
 
-    for( dv_i=poMDTable->attr_begin(); dv_i != poMDTable->attr_end(); dv_i++ )
+   for( dv_i=poMDTable->attr_begin(); 
+		 dv_i != poMDTable->attr_end(); dv_i++ )
     {
-        if( poMDTable->get_attr_type( dv_i ) != Attr_string )
-            continue;
-
-        poTarget->SetMetadataItem( 
+		poTarget->SetMetadataItem( 
             poMDTable->get_name( dv_i ).c_str(), 
-            StripQuotes( poMDTable->get_attr(dv_i) ).c_str() );
+            poMDTable->get_attr( dv_i ).c_str() );
     }
+
+/* -------------------------------------------------------------------- */
+/*      Add band imformation to each band metadata.                     */
+/*	    if the dimension of the band is three and more,                 */
+/*	    the dimension name and its number                               */
+/*      are added in the band metadata.                                 */
+/*	    e.g. if containerName is like                                   */
+/*          "[3][12][yc][xc]?[time][depth][latitude][lonongitude]",     */
+/*	        then "dods_coordinates" and "dods_dimensions"               */
+/*	        are "[3][12]" and "[time][depth]" respectively.             */
+/* -------------------------------------------------------------------- */
+	// set variable name of each band
+    poTarget->SetMetadataItem( "dods_variable", containerName.c_str() );
+
+	string::size_type index;
+	string::size_type index2;
+	if ( coordName.c_str() != NULL )
+	{
+		index = coordName.find('?');
+		if( index != string::npos )
+		{
+			string tmpSubString = coordName.substr(0,index);
+			for (int i=0; i<2; i++)
+			{
+				index2 = tmpSubString.find_last_of('[');
+				if( index2 != string::npos )
+					tmpSubString = tmpSubString.substr(0,index2);
+			}
+			if (tmpSubString != "")
+			{
+				poTarget->SetMetadataItem( "dods_coordinates", 
+										tmpSubString.c_str() );
+			}
+
+			tmpSubString = coordName.substr(index+1);
+			for (int i=0; i<2; i++)
+			{
+				index2 = tmpSubString.find_last_of('[');
+				if( index2 != string::npos )
+					tmpSubString = tmpSubString.substr(0,index2);
+			}
+			if (tmpSubString != "")
+			{
+				poTarget->SetMetadataItem( "dods_dimensions", 
+										tmpSubString.c_str() );
+			}
+		}
+	}
 }
 
 /************************************************************************/
 /*                             HarvestDAS()                             */
 /************************************************************************/
 
-void DODSDataset::HarvestDAS()
+void DODSDataset::HarvestDAS(string gridMapping, string gridNameX, string gridNameY)
 
 {
 /* -------------------------------------------------------------------- */
 /*      Try and fetch the corresponding DAS subtree if it exists.       */
+/*        NB : If grid_mapping is given and it does not include			*/
+/*				"spatial_ref", "proj4_string" or "Geotransform"			*/
+/*				the information in grid_mapping is lost.				*/
+/*			   But it might be good to take "grid_mapping_name",        */
+/*				if it exists.	                                        */
 /* -------------------------------------------------------------------- */
+	string GlobalMetadata;
+    DDS::Vars_iter v_i;
+
 #ifdef LIBDAP_39
-    AttrTable *poFileInfo = oDAS.get_table( "GLOBAL" );
+	AttrTable *poGridInfo = oDAS.get_table( gridMapping.c_str() );
+	AttrTable *poFileInfo = oDAS.get_table( "GLOBAL" );
+	GlobalMetadata = "GLOBAL";
 
     if( poFileInfo == NULL )
     {
         poFileInfo = oDAS.get_table( "NC_GLOBAL" );
+		GlobalMetadata = "NC_GLOBAL";
 
 	if( poFileInfo == NULL )
 	{
 	    poFileInfo = oDAS.get_table( "HDF_GLOBAL" );
+		GlobalMetadata = "HDF_GLOBAL";
 
 	    if( poFileInfo == NULL )
 	    {
@@ -721,15 +901,19 @@
 	}
     }
 #else
+	AttrTable *poGridInfo = oDAS.find_container( gridMapping.c_str() );
     AttrTable *poFileInfo = oDAS.find_container( "GLOBAL" );
+	GlobalMetadata = "GLOBAL";
 
     if( poFileInfo == NULL )
     {
         poFileInfo = oDAS.find_container( "NC_GLOBAL" );
+		GlobalMetadata = "NC_GLOBAL";
 
 	if( poFileInfo == NULL )
 	{
 	    poFileInfo = oDAS.find_container( "HDF_GLOBAL" );
+		GlobalMetadata = "HDF_GLOBAL";
 
 	    if( poFileInfo == NULL )
 	    {
@@ -741,74 +925,103 @@
 #endif
 
 /* -------------------------------------------------------------------- */
-/*      Try and fetch the bounds                                        */
+/*      Try and fetch a GeoTransform in the dataset                     */
+/*      if it is not computed from arrays in HarvestMaps()              */
+/*      This allows us to represent rotated and sheared images.         */
 /* -------------------------------------------------------------------- */
-    string oNorth, oSouth, oEast, oWest;
+	if (!bGotGeoTransform)
+	{
+		string oValue;
 
-    oNorth = poFileInfo->get_attr( "Northernmost_Northing" );
-    oSouth = poFileInfo->get_attr( "Southernmost_Northing" );
-    oEast = poFileInfo->get_attr( "Easternmost_Easting" );
-    oWest = poFileInfo->get_attr( "Westernmost_Easting" );
+		oValue = poFileInfo->get_attr( "GeoTransform" );
+			if( oValue == "" )
+				oValue = poGridInfo->get_attr( "GeoTransform" );
+		if( oValue != "" )
+		{
+			char **papszItems = CSLTokenizeString( oValue.c_str() );
+			if( CSLCount(papszItems) == 6 )
+			{
+				adfGeoTransform[0] = atof(papszItems[0]);
+				adfGeoTransform[1] = atof(papszItems[1]);
+				adfGeoTransform[2] = atof(papszItems[2]);
+				adfGeoTransform[3] = atof(papszItems[3]);
+				adfGeoTransform[4] = atof(papszItems[4]);
+				adfGeoTransform[5] = atof(papszItems[5]);
+				bGotGeoTransform = TRUE;
+			}
+			else
+			{
+				CPLError( CE_Warning, CPLE_AppDefined, 
+						  "Failed to parse GeoTransform DAS value: %s",
+						  oValue.c_str() );
+			}
+			CSLDestroy( papszItems );
+		}
+	}
 
-    if( oNorth != "" && oSouth != "" && oEast != "" && oWest != "" )
-    {
-        adfGeoTransform[0] = atof(oWest.c_str());
-        adfGeoTransform[1] = 
-            (atof(oEast.c_str()) - atof(oWest.c_str())) / nRasterXSize;
-        adfGeoTransform[2] = 0.0;
-        adfGeoTransform[3] = atof(oNorth.c_str());
-        adfGeoTransform[4] = 0.0;
-        adfGeoTransform[5] = 
-            (atof(oSouth.c_str()) - atof(oNorth.c_str())) / nRasterYSize;
-
-        bGotGeoTransform = TRUE;
-    }
-
 /* -------------------------------------------------------------------- */
-/*      Try and fetch a GeoTransform.  The result will override the     */
-/*      geotransform derived from the bounds if it is present.  This    */
-/*      allows us to represent rotated and sheared images.              */
+/*      if a GeoTransform is not given from the dataset or              */
+/*      from arrays in HarvestMaps(),                                   */
+/*      try and fetch the bounds from the bounds and compute            */
+/*      adfGeoTransform from them.                                      */
 /* -------------------------------------------------------------------- */
-    string oValue;
 
-    oValue = StripQuotes(poFileInfo->get_attr( "GeoTransform" ));
-    if( oValue != "" )
-    {
-        char **papszItems = CSLTokenizeString( oValue.c_str() );
-        if( CSLCount(papszItems) == 6 )
-        {
-            adfGeoTransform[0] = atof(papszItems[0]);
-            adfGeoTransform[1] = atof(papszItems[1]);
-            adfGeoTransform[2] = atof(papszItems[2]);
-            adfGeoTransform[3] = atof(papszItems[3]);
-            adfGeoTransform[4] = atof(papszItems[4]);
-            adfGeoTransform[5] = atof(papszItems[5]);
-            bGotGeoTransform = TRUE;
-        }
-        else
-        {
-            CPLError( CE_Warning, CPLE_AppDefined, 
-                      "Failed to parse GeoTransform DAS value: %s",
-                      oValue.c_str() );
-        }
-        CSLDestroy( papszItems );
-    }
+	if (!bGotGeoTransform)
+	{
+		string oNorth, oSouth, oEast, oWest;
 
+		oNorth = poFileInfo->get_attr( "Northernmost_Northing" );
+		oSouth = poFileInfo->get_attr( "Southernmost_Northing" );
+		oEast = poFileInfo->get_attr( "Easternmost_Easting" );
+		oWest = poFileInfo->get_attr( "Westernmost_Easting" );
+
+		if( oNorth == "" || oSouth == "" || oEast == "" || oWest == "" )
+		{
+			oNorth = poGridInfo->get_attr( "Northernmost_Northing" );
+			oSouth = poGridInfo->get_attr( "Southernmost_Northing" );
+			oEast = poGridInfo->get_attr( "Easternmost_Easting" );
+			oWest = poGridInfo->get_attr( "Westernmost_Easting" );
+		}
+
+		if( oNorth != "" && oSouth != "" && oEast != "" && oWest != "" )
+		{
+			adfGeoTransform[0] = atof(oWest.c_str());
+			adfGeoTransform[1] = 
+				(atof(oEast.c_str()) - atof(oWest.c_str())) / nRasterXSize;
+			adfGeoTransform[2] = 0.0;
+			adfGeoTransform[3] = atof(oNorth.c_str());
+			adfGeoTransform[4] = 0.0;
+			adfGeoTransform[5] = 
+				(atof(oSouth.c_str()) - atof(oNorth.c_str())) / nRasterYSize;
+		}
+	}
+
 /* -------------------------------------------------------------------- */
 /*      Get the Projection.  If it doesn't look like "pure" WKT then    */
 /*      try to process it through SetFromUserInput().  This expands     */
 /*      stuff like "WGS84".                                             */
 /* -------------------------------------------------------------------- */
-    oWKT = StripQuotes(poFileInfo->get_attr( "spatial_ref" ));
-    // strip remaining backslashes (2007-04-26, gaffigan@sfos.uaf.edu)
+	string unitsX, unitsY, units;
+	AttrTable *poGridX, *poGridY;
+
+	oWKT = poFileInfo->get_attr( "spatial_ref" );
+	if (oWKT == "")
+		oWKT = poFileInfo->get_attr( "proj4_string" );
+	if (oWKT == "")
+		oWKT = poGridInfo->get_attr( "spatial_ref" );
+	if (oWKT == "")
+		oWKT = poGridInfo->get_attr( "proj4_string" );
+
+	// strip remaining backslashes (2007-04-26, gaffigan@sfos.uaf.edu)
     oWKT.erase(std::remove(oWKT.begin(), oWKT.end(), '\\'), oWKT.end());
+
+	// check if check is WKT or otehrs (e.g. proj4) 
     if( oWKT.length() > 0 
         && !EQUALN(oWKT.c_str(),"GEOGCS",6)
         && !EQUALN(oWKT.c_str(),"PROJCS",6)
         && !EQUALN(oWKT.c_str(),"LOCAL_CS",8) )
     {
         OGRSpatialReference oSRS;
-
         if( oSRS.SetFromUserInput( oWKT.c_str() ) != OGRERR_NONE )
         {
             CPLError( CE_Warning, CPLE_AppDefined, 
@@ -819,16 +1032,49 @@
         else
         {
             char *pszProcessedWKT = NULL;
-            oSRS.exportToWkt( &pszProcessedWKT );
-            oWKT = pszProcessedWKT;
+/* -------------------------------------------------------------------- */
+/*			if units are given in x and y grid metadata,				*/
+/*          add it to oSRS												*/
+/* -------------------------------------------------------------------- */
+			poGridX = oDAS.get_table( gridNameX.c_str() );
+			if (poGridX != NULL)
+				unitsX = poGridX->get_attr("units");
+			poGridY = oDAS.get_table( gridNameY.c_str() );
+			if (poGridY != NULL)
+				unitsY = poGridY->get_attr("units");
+
+			/* TODO: what to do if units are not equal in X and Y */
+            
+			if ((unitsX != "") && (unitsY != "") && (unitsX == unitsY))
+				units = unitsX;
+
+            /* add units to PROJCS */
+            if ( units != "" ) 
+			{
+                if ( units == "m" ) {
+                    oSRS.SetLinearUnits( "metre", 1.0 );
+                    oSRS.SetAuthority( "PROJCS|UNIT", "EPSG", 9001 );
+                }
+                else if ( units == "km" ) {
+                    oSRS.SetLinearUnits( "kilometre", 1000.0 );
+                    oSRS.SetAuthority( "PROJCS|UNIT", "EPSG", 9036 );
+                }
+                /* TODO check for other values */
+                // else 
+                //     oSRS.SetLinearUnits(pszUnits, 1.0);
+            }
+
+            oSRS.exportToWkt( &pszProcessedWKT );			
+			oWKT = pszProcessedWKT;
             CPLFree( pszProcessedWKT );
         }
+
     }
-
 /* -------------------------------------------------------------------- */
 /*      Collect Metadata.                                               */
 /* -------------------------------------------------------------------- */
-    HarvestMetadata( this, poFileInfo );
+	string option;
+    HarvestMetadata( this, poFileInfo, GlobalMetadata, option);
 }
 
 /************************************************************************/
@@ -861,8 +1107,8 @@
          iterMap != poGrid->map_end(); 
          iterMap++, iMap++ )
     {
-        if( iMap == iXDim )
-            poXMap = dynamic_cast<Array *>(*iterMap);
+        if( iMap == iXDim  )
+			poXMap = dynamic_cast<Array *>(*iterMap);			
         else if( iMap == iYDim )
             poYMap = dynamic_cast<Array *>(*iterMap);
     }
@@ -870,13 +1116,15 @@
     if( poXMap == NULL || poYMap == NULL )
         return;
 
-    if( poXMap->var()->type() != dods_float64_c 
-        || poYMap->var()->type() != dods_float64_c )
+    if(( poXMap->var()->type() != dods_float64_c 
+		||  poYMap->var()->type() != dods_float64_c) 
+		&& ( poXMap->var()->type() != dods_float32_c 
+		||  poYMap->var()->type() != dods_float32_c))
     {
         CPLDebug( "DODS", "Ignoring Grid Map - not a supported data type." );
         return;							       
     }
-
+	
 /* -------------------------------------------------------------------- */
 /*      TODO: We ought to validate the dimension of the map against our */
 /*      expected size.                                                  */
@@ -892,25 +1140,41 @@
 /*      Build constraint expression.                                    */
 /* -------------------------------------------------------------------- */
     string constraint;
-    string xdimname = "lon";
-    string ydimname = "lat";
-    
+	string xdimname;
+    string ydimname;
+	string tmpString;
+	string::size_type index, indexStart, len;    
 
-    constraint = oVarName + "." + xdimname + "," 
-        + oVarName + "." + ydimname;
+	index = oCE.find("?");
+    if( index != string::npos )
+		tmpString = oCE.substr(index+1);
+	else
+		tmpString = oCE;
 
+	len = tmpString.length();
+	indexStart = tmpString.find_last_of('[');
+	if( indexStart != string::npos )
+		xdimname = tmpString.substr(indexStart+1, len-indexStart-2 );
+
+	tmpString = tmpString.substr(0,indexStart);
+	len = tmpString.length();
+	indexStart = tmpString.find_last_of('[');
+	if( indexStart != string::npos )
+		ydimname = tmpString.substr(indexStart+1, len-indexStart-2 );
+
+	constraint = xdimname + "," + ydimname;
 /* -------------------------------------------------------------------- */
 /*      Request data from server.                                       */
 /* -------------------------------------------------------------------- */
-    DataDDS data( poBaseTypeFactory );
-    
-    GetConnect()->request_data(data, constraint );
+	DataDDS data( poBaseTypeFactory );
+	GetConnect()->request_data(data, constraint );
 
 /* -------------------------------------------------------------------- */
 /*      Get the DataDDS Array object from the response.                 */
 /* -------------------------------------------------------------------- */
-    BaseType *poBtX = get_variable(data, oVarName + "." + xdimname );
-    BaseType *poBtY = get_variable(data, oVarName + "." + ydimname );
+	BaseType *poBtX = get_variable(data, xdimname );
+    BaseType *poBtY = get_variable(data, ydimname );
+
     if (!poBtX || !poBtY 
         || poBtX->type() != dods_array_c 
         || poBtY->type() != dods_array_c )
@@ -948,21 +1212,21 @@
 /*      assuming the maps are linear and refer to the center of the     */
 /*      pixels.                                                         */
 /* -------------------------------------------------------------------- */
-    bGotGeoTransform = TRUE;
-
-    // pixel size. 
+	bGotGeoTransform = TRUE;
+    // pixel size 
     adfGeoTransform[1] = (padfXMap[nRasterXSize-1] - padfXMap[0])
-        / (nRasterXSize-1);
+							/ (nRasterXSize-1);
     adfGeoTransform[5] = (padfYMap[nRasterYSize-1] - padfYMap[0])
-        / (nRasterYSize-1);
+							/ (nRasterYSize-1);
+	
+	// origin
+	adfGeoTransform[0] = padfXMap[0] - adfGeoTransform[1] * 0.5;
+	adfGeoTransform[3] = padfYMap[0] - adfGeoTransform[5] * 0.5;
 
-    // rotational coefficients. 
-    adfGeoTransform[2] = 0.0;
-    adfGeoTransform[4] = 0.0;
+	// rotational coefficients
+	adfGeoTransform[2] = 0.0;
+	adfGeoTransform[4] = 0.0;
 
-    // origin/ 
-    adfGeoTransform[0] = padfXMap[0] - adfGeoTransform[1] * 0.5;
-    adfGeoTransform[3] = padfYMap[0] - adfGeoTransform[5] * 0.5;
 }
 
 /************************************************************************/
@@ -975,7 +1239,6 @@
     if( !EQUALN(poOpenInfo->pszFilename,"http://",7) 
         && !EQUALN(poOpenInfo->pszFilename,"https://",8) )
         return NULL;
-
     
     DODSDataset *poDS = new DODSDataset();
     char **papszVarConstraintList = NULL;
@@ -1002,7 +1265,7 @@
             poDS->oURL = oWholeName.substr(0,t_char);
             oVarList = oWholeName.substr(t_char+1,oWholeName.length());
         }
-
+				
 /* -------------------------------------------------------------------- */
 /*      Get the AISConnect instance and the DAS and DDS for this        */
 /*      server.                                                         */
@@ -1017,7 +1280,6 @@
 /*      DDS and try to identify grids or arrays that are good           */
 /*      targets and return them in the same format.                     */
 /* -------------------------------------------------------------------- */
-
         if( oVarList.length() == 0 )
             papszVarConstraintList = poDS->CollectBandsFromDDS();
         else
@@ -1032,52 +1294,78 @@
 /* -------------------------------------------------------------------- */
 /*      For now we support only a single band.                          */
 /* -------------------------------------------------------------------- */
-        DODSRasterBand *poBaseBand = 
+		string gridName;
+		DODSRasterBand *poBaseBand = 
             new DODSRasterBand(poDS, 
                                string(papszVarConstraintList[0]),
                                string(papszVarConstraintList[1]), 
                                1 );
-
         poDS->nRasterXSize = poBaseBand->GetXSize();
         poDS->nRasterYSize = poBaseBand->GetYSize();
             
-	poDS->SetBand(1, poBaseBand );
+		poDS->SetBand(1, poBaseBand );
 
+		// if grid_mapping is given in the baseband, take the value
+		if (poBaseBand->GetMetadataItem( "grid_mapping" ) != NULL)
+			gridName = poBaseBand->GetMetadataItem( "grid_mapping" );
+
         for( int iBand = 1; papszVarConstraintList[iBand*2] != NULL; iBand++ )
         {
-            poDS->SetBand( iBand+1, 
-                           new DODSRasterBand(poDS, 
+			DODSRasterBand *poiBand = 
+				new DODSRasterBand(poDS, 
                                    string(papszVarConstraintList[iBand*2+0]),
                                    string(papszVarConstraintList[iBand*2+1]), 
-                                              1 ) );
+                                   1 );
+			poDS->SetBand(iBand+1, poiBand );
+
+			if (gridName == "" && 
+				poiBand->GetMetadataItem( "grid_mapping" ) != NULL)
+				gridName = poiBand->GetMetadataItem( "grid_mapping" );
         }
 
 /* -------------------------------------------------------------------- */
+/*      Look for "map" information for a grid.                          */
+/* -------------------------------------------------------------------- */
+		poDS->HarvestMaps( string(papszVarConstraintList[0]),
+                               string(papszVarConstraintList[1]) );
+				
+/* -------------------------------------------------------------------- */
+/*      get grid X and Y names from the first band in VarConstraintList */
+/* -------------------------------------------------------------------- */
+		string gridNameX, gridNameY;
+		string oCE = papszVarConstraintList[1];
+		int index, indexStart, len;
+
+		index = oCE.find("?");
+		string tmpString = oCE.substr(index+1);
+
+		len = tmpString.length();
+		indexStart = tmpString.find_last_of('[');
+		if( indexStart != string::npos )
+			gridNameX = tmpString.substr(indexStart+1, len-indexStart-2 );
+
+		tmpString = tmpString.substr(0,indexStart);
+		len = tmpString.length();
+		indexStart = tmpString.find_last_of('[');
+		if( indexStart != string::npos )
+			gridNameY = tmpString.substr(indexStart+1, len-indexStart-2 );
+
+/* -------------------------------------------------------------------- */
 /*      Harvest DAS dataset level information including                 */
 /*      georeferencing, and metadata.                                   */
 /* -------------------------------------------------------------------- */
-        poDS->HarvestDAS();
+		poDS->HarvestDAS( gridName, gridNameX, gridNameY);
 
-/* -------------------------------------------------------------------- */
-/*      If we don't have georeferencing, look for "map" information     */
-/*      for a grid.                                                     */
-/* -------------------------------------------------------------------- */
-        if( !poDS->bGotGeoTransform )
-        {
-            poDS->HarvestMaps( string(papszVarConstraintList[0]),
-                               string(papszVarConstraintList[1]) );
-        }
     }
 
     catch (Error &e) {
         string msg =
-"An error occurred while creating a virtual connection to the DAP server:\n";
+			"An error occurred while creating a virtual connection to the DAP server:\n";
         msg += e.get_error_message();
         CPLError(CE_Failure, CPLE_AppDefined, "%s", msg.c_str());
         delete poDS;
         poDS = NULL;
     }
-
     CSLDestroy(papszVarConstraintList);
     
 /* -------------------------------------------------------------------- */
@@ -1091,7 +1379,6 @@
                   " datasets.\n" );
         return NULL;
     }
-    
     return poDS;
 }
 
@@ -1128,23 +1415,40 @@
 /************************************************************************/
 
 DODSRasterBand::DODSRasterBand(DODSDataset *poDSIn, string oVarNameIn, 
-                               string oCEIn, int nOverviewFactorIn )
+                               string oConstIn, int nOverviewFactorIn )
 {
+	
     poDS = poDSIn;
 
     bTranspose = FALSE;
     bFlipX = FALSE;
     bFlipY = FALSE;
-
     oVarName = oVarNameIn;
-    oCE = oCEIn;
     nOverviewFactor = nOverviewFactorIn;
     eColorInterp = GCI_Undefined;
     poCT = NULL;
-
     nOverviewCount = 0;
     papoOverviewBand = NULL;
+	string oCEIn;
+	string::size_type index;    
 
+/*----------------------------------------------------------------------*/
+/*	if oConstIn has not only "dimension of variable"                    */
+/*	but also "variableName" like "[0][y][x]?[time][yc][xc]"             */
+/*	then take only "dimension of variable" ([0][y][x])                  */
+/*   and set it to oCEIn												*/
+/*----------------------------------------------------------------------*/
+	index = oConstIn.find("?");
+    if( index == string::npos )
+    {
+		oCEIn = oConstIn;
+    }
+    else
+    {
+		oCEIn = oConstIn.substr(0,index);
+    }
+	oCE = oCEIn;
+
 /* -------------------------------------------------------------------- */
 /*      Fetch the DDS definition, and isolate the Array.                */
 /* -------------------------------------------------------------------- */
@@ -1201,9 +1505,8 @@
     {
         throw Error("Variable does not have even 2 dimensions.  For now this is required." );
     }
-
     int nXDir = 1, nYDir = 1;
-    int iXDim = GetDimension( oCE, "x", &nXDir );
+	int iXDim = GetDimension( oCE, "x", &nXDir );
     int iYDim = GetDimension( oCE, "y", &nYDir );
 
     if( iXDim == -1 || iYDim == -1 )
@@ -1213,12 +1516,9 @@
 
     Array::Dim_iter x_dim = poArray->dim_begin() + iXDim;
     Array::Dim_iter y_dim = poArray->dim_begin() + iYDim;
-
     nRasterXSize = poArray->dimension_size( x_dim ) / nOverviewFactor;
     nRasterYSize = poArray->dimension_size( y_dim ) / nOverviewFactor;
-
     bTranspose = iXDim < iYDim;
-
     bFlipX = (nXDir == -1);
     bFlipY = (nYDir == -1);
 
@@ -1261,7 +1561,7 @@
 /*      Get other information from the DAS for this band.               */
 /* -------------------------------------------------------------------- */
     if( nOverviewFactorIn == 1 )
-        HarvestDAS();
+        HarvestDAS(oConstIn);
 
 /* -------------------------------------------------------------------- */
 /*      Create overview band objects.                                   */
@@ -1311,7 +1611,7 @@
 /*                             HarvestDAS()                             */
 /************************************************************************/
 
-void DODSRasterBand::HarvestDAS()
+void DODSRasterBand::HarvestDAS(string dimName)
 
 {
     DODSDataset *poDODS = dynamic_cast<DODSDataset *>(poDS);
@@ -1319,6 +1619,7 @@
 /* -------------------------------------------------------------------- */
 /*      Try and fetch the corresponding DAS subtree if it exists.       */
 /* -------------------------------------------------------------------- */
+
 #ifdef LIBDAP_39
     AttrTable *poBandInfo = poDODS->GetDAS().get_table( oVarName );
 #else
@@ -1334,7 +1635,7 @@
 /* -------------------------------------------------------------------- */
 /*      collect metadata.                                               */
 /* -------------------------------------------------------------------- */
-    poDODS->HarvestMetadata( this, poBandInfo );
+    poDODS->HarvestMetadata( this, poBandInfo, oVarName, dimName);
 
 /* -------------------------------------------------------------------- */
 /*      Get photometric interpretation.                                 */
@@ -1375,12 +1676,10 @@
 	}
     }
 
-
-
 /* -------------------------------------------------------------------- */
 /*      Collect color table                                             */
 /* -------------------------------------------------------------------- */
-    AttrTable *poCTable = poBandInfo->find_container("Colormap");
+    AttrTable *poCTable = poBandInfo->find_container( "Colormap" );
     if( poCTable != NULL )
     {
         AttrTable::Attr_iter dv_i;
@@ -1431,7 +1730,7 @@
 CPLErr 
 DODSRasterBand::IReadBlock(int nBlockXOff, int nBlockYOff, void *pImage)
 {
-    DODSDataset *poDODS = dynamic_cast<DODSDataset *>(poDS);
+	DODSDataset *poDODS = dynamic_cast<DODSDataset *>(poDS);
     int nBytesPerPixel = GDALGetDataTypeSize(eDataType) / 8;
 
 /* -------------------------------------------------------------------- */
@@ -1489,7 +1788,6 @@
 /* -------------------------------------------------------------------- */
     try {
         DataDDS data( poDODS->poBaseTypeFactory );
-
         poDODS->GetConnect()->request_data(data, final_constraint );
 
 /* -------------------------------------------------------------------- */
@@ -1734,9 +2032,45 @@
         poDriver->SetMetadataItem( GDAL_DMD_HELPTOPIC, 
                                    "frmt_various.html#DODS" );
 
+
         poDriver->pfnOpen = DODSDataset::Open;
 
         GetGDALDriverManager()->RegisterDriver( poDriver );
     }
 }
 
+/************************************************************************/
+/*                         Direct_Product()                             */
+/************************************************************************/
+char **Direct_Product(int dimension, char **oConst)
+{
+/* -------------------------------------------------------------------- */
+/*	 Create all combinations based on dimension and oConst.             */
+/*	 e.g.) if dimension=3 and oConst[0]=[yc][xc],                       */
+/*			then create [0][yc][xc], [1][yc][xc], [2][yc][xc] and       */
+/*				 set them into **papszCombinationList                   */
+/*		  if dimension=4 and oConst[0]=[0][y][x],					    */
+/*							 oConst[1]=[1][y][x], oConst[2]=[2][y][x]   */
+/*			then create [0][0][y][x], [1][0][y][x], [2][0][y][x],       */
+/*						[3][0][y][x], [0][1][y][x], ..., [3][2][y][x]   */
+/*				 set them into **papszCombinationList                   */
+/* -------------------------------------------------------------------- */
+	char **papszCombinationList = NULL;	
+	char chr[3];
+
+	for (int i=0; oConst[i] != NULL ; i++)
+	{
+		for (int j=0; j<dimension; j++)
+		{
+			sprintf(chr, "%d", j );
+			string combination = "[";
+			combination += string( chr );
+			combination += "]";
+			combination += oConst[i];
+			papszCombinationList = CSLAddString( papszCombinationList, combination.c_str() );	
+		}
+	}
+	return papszCombinationList;
+}
+
+
